#include <vector>
#include <Eigen/Sparse>

#ifndef ELEMENT
#define ELEMENT

using namespace Eigen;

//! Element contains element data and relevant functions
class Element
{

public:
	EIGEN_MAKE_ALIGNED_OPERATOR_NEW		//!< Neccessary for std::vector of objects with Eigen members.

	Element();
	int i;		//!< Node 1 in the element
	int	j;		//!< Node 2 in the element
	int k;		//!< Node 3 in the element

	double area;		//!< Elements total area
	double material;	//!< Young's modulus for the element
	double shearModule;

	double Iex;		//!< Elements second area moment x
	double Iey;		//!< Elements second area moment y
	double Iexy;	//!< Elements second area moment product

	double xl1, xl2, xl3;	//!< Distance in x direction from node to element area center.
	double yl1, yl2, yl3;	//!< Distance in y direction from node to element area center.

	double ecx, ecy;		//!< Elements area centre
	double dx, dy;			//!< Distance between mesh- and node area centre

	double x1, x2, x3;		//!< Elements nodepositions in x-directions
	double y1, y2, y3;		//!< Elements nodepositions in x-directions

	double xp1, xp2, xp3;	//!< Nodal x-coordinates in principal axes
	double yp1, yp2, yp3;	//!< Nodal y-coordinates in principal axes

	/*!
		Returns the x-coordinate of a point in the element with respect to its area center as a function of area coordinates. \n
		\f$ x = (x_{l1} - x_{l3})\zeta_1 + (x_{l2} - x_{l3})\zeta_2 + x_{l3} \f$
	*/
	
	double xOfZeta(double zeta1, double zeta2);

	/*!
			Returns the x- and y-coordinates of a point in the element with respect to its area center as a function of area coordinates. \n
		\f$ y = (y_{l1} - y_{l3})\zeta_1 + (y_{l2} - y_{l3})\zeta_2 + y_{l3} \f$
	*/

	double yOfZeta(double zeta1, double zeta2);

	/*!
		Calculates the area of the element. \n
		\f$ A_e = \frac{1}{2} \cdot |J| \f$
	*/
	void calculateArea();

	/*!
		Calculates the area centre of the element. \n
		\f$ ec_x = \frac{x1 + x2 + x3}{3} \f$ \n
		\f$ ec_y = \frac{y1 + y2 + y3}{3} \f$
	*/
	void calculateAreaCentre();

	/*!
		Calculates the second moment of area for the element.
		\param Ay Area centre of the mesh

		\f$ dy_e = A_y - Ac_{ey} \f$ \n
		\f$ I_x = \sum{( I_{x_e} + dy_e^2*A_e) } \f$ 
	*/
	double getSecondAreaMomentX(double);
	
	/*!
		Calculates the second moment of area for the element.
		\param Ax Area centre of the mesh

		\f$ dx_e = A_x - Ac_{ex} \f$ \n
		\f$ I_y =  \sum{(I_{y_e} + dx_e^2*A_e)} \f$ 
	*/
	double getSecondAreaMomentY(double);

	/*!
		Calculating the product of moment of inertia

		\f$ dx_e = Ac_x - Ac_{ex} \f$ \n
		\f$ dy_e = Ac_y - Ac_{ey} \f$ \n
		\f$ I_{xy} = \sum{(I_{xy_e} + dx_e*dy_e*A_e)} \f$ 
	*/
	double getSecondAreaMomentProduct(double, double);

	/*!
		Calculates the Matrix containing shape fucntions \n

		\return MatrixXd(1,3) with shape functions
	*/
	MatrixXd getShapefunction(double zeta1, double zeta2, double zeta3);

	/*!
		Returns the shapefunction as a function of x and y.
		\f$ \zeta_1 = y_{23}x + x_{32}y + (x_2y_3 - x_3y_2) \f$ \n
		\f$ \zeta_1 = y_{31}x + x_{13}y + (x_3y_1 - x_1y_3) \f$ \n
		\f$ \zeta_1 = y_{12}x + x_{21}y + (x_1y_2 - x_2y_1) \f$

		\return MatrixXd(1, 3) with shape function.
	*/

	MatrixXd getShapefunction(double x, double y);

	MatrixXd getNMatrix(double zeta1, double zeta2, double zeta3); //!< Returns a matrix of matrices with shape functions
	MatrixXd getNMatrix(double x, double y);	

	MatrixXd B;	//!< 2X3 Matrix containing differentials of shapefunction with respect to x and y respectively
	MatrixXd G; //!< Matrix with shearmodule used to determine initial element loads in tortional analysis
	MatrixXd XoverY;	//!< 6x1 matrix with x1, x2, x3, y1, y2 and y3 of the element.
	MatrixXd xCoordinates;	//!< 1x3 matrix with x-coordinates.
	MatrixXd yCoordinates;	//!< 1X3 matrix with y-coordinates.

	void calculateConstants();	//!< Calculates the constant element matrices

	//Vector2d shearStress(int n); //!< Returns nodal shear stress due to tortional moment in node n of the element

	MatrixXd elementStiffness;	//!< Matrix used to temporarily hold element stiffnes coefficients.
	MatrixXd elementLoad;	//!< Matrix used to temporarily hold element load coefficients.
	MatrixXd elementDisplacement;	//!< Matrix used to temporarily hold element displacement values.
	
	MatrixXd initialStrain;	//!< Matrix used to temporarily hold element initial strain coefficients.
	
	/*!
		Creates a vector with node numbers that makes up the element.
		\return Vector<Int> containing element nodes.
	*/
	std::vector<int> getIEG();

	void calculateStiffnessMatrix();	// !< Calculates the element stiffness matrix (k).

	//void calculateElementLoad();	//!<Calculates the initial element load. (Torsional moment about beam axis.)
	void calculateElementLoadTorsion(); //!<Calculates the initial load vector for torsion.
	void calculateElementLoadShearX(); //!<Calculates the initial load vector for shear along X.
	void calculateElementLoadShearY(); //!<Calculates the initial load vector for shear along Y.


	/*!
		Uses getIEG() to assign system values to corresponding local parameters.
	*/

	void setElementValue(MatrixXd &system, MatrixXd &local);

	std::vector<MatrixXd> tau;	//!< Vector temporarily storing shear stresses during calculation.

	/*!
		Calculates shear stresses at nodes using initialStrain and elementsDisplacement\n
		\f$ \tau = Bv + \epsilon_0 \f$ 
	*/

	void computeTau();	

private:
};

#endif