#include "stdafx.h"
#include "math.h"
#include "Mesh.h"
#include <vector>
#include <map>
#include <algorithm>

using namespace Eigen;

Mesh::Mesh(std::vector<Element> elements, int nElem, std::vector<Node> nodes, int nNode, double material)
{
	Mesh::elements = elements;
	Mesh::nElem = nElem;
	Mesh::nodes = nodes;
	Mesh::nNode = nNode;

	for(unsigned e=0; e<elements.size(); e++){
		std::vector<int> IEG = elements[e].getIEG();
		for(int i=0; i<3; i++){
			Mesh::nodes[IEG[i]].depElem.push_back(e);
			Mesh::nodes[IEG[i]].depIEG.push_back(i);
		}
	}


	setEModule(material);

	calculateMeshProperties();

	transformCoordinates();

	solveBendingMoments();

	solveTorsion();
	flush();
	
	/*solveShearX();
	flush();
	solveShearY();
	flush();*/

}

Mesh::Mesh(){};

// ********** Mesh Properties **********

void Mesh::setEModule(double mat)
{
	E = mat;

	for(unsigned e=0; e<elements.size(); e++){
		elements[e].material = mat;
	}

}

void Mesh::calculateMeshProperties()
{
	double Sx = 0, Sy = 0;
	Ix = 0; 
	Iy = 0;
	totalArea = 0;
	areaStiffness = 0;
	
	for (int e = 0; e < nElem; e++){

		elements[e].calculateAreaCentre();
		Sx += elements[e].area * elements[e].ecx * elements[e].material;
		Sy += elements[e].area * elements[e].ecy * elements[e].material;
		totalArea += elements[e].area;
		areaStiffness += elements[e].area * elements[e].material;
	}
	Ax = Sx/areaStiffness;
	Ay = Sy/areaStiffness;

	sigmaAxial.resize(nodes.size(), 1);
	for(unsigned n=0; n<nodes.size(); n++){
		sigmaAxial(n, 0) = 1.0/totalArea;
	}

	
	calculateSecondAreaMoments();
}

void Mesh::transformCoordinates(){
	if(anglePrincipalAxes!=0 || Ax!=0 || Ay!=0){
		for(unsigned e=0; e<elements.size(); e++){
			elements[e].xp1 = (elements[e].x1-Ax) * cos(anglePrincipalAxesRad) + (elements[e].y1-Ay) * sin(anglePrincipalAxesRad);
			elements[e].xp2 = (elements[e].x2-Ax) * cos(anglePrincipalAxesRad) + (elements[e].y2-Ay) * sin(anglePrincipalAxesRad);
			elements[e].xp3 = (elements[e].x3-Ax) * cos(anglePrincipalAxesRad) + (elements[e].y3-Ay) * sin(anglePrincipalAxesRad);

			elements[e].yp1 = (elements[e].y1-Ay) * cos(anglePrincipalAxesRad) - (elements[e].x1-Ax) * sin(anglePrincipalAxesRad);
			elements[e].yp2 = (elements[e].y2-Ay) * cos(anglePrincipalAxesRad) - (elements[e].x2-Ax) * sin(anglePrincipalAxesRad);
			elements[e].yp3 = (elements[e].y3-Ay) * cos(anglePrincipalAxesRad) - (elements[e].x3-Ax) * sin(anglePrincipalAxesRad);
			elements[e].x1 = elements[e].xp1;
			elements[e].x2 = elements[e].xp2;
			elements[e].x3 = elements[e].xp3;
			elements[e].y1 = elements[e].yp1;
			elements[e].y2 = elements[e].yp2;
			elements[e].y3 = elements[e].yp3;

			elements[e].calculateConstants();
		}
		for(unsigned n=0; n<nodes.size(); n++){

			nodes[n].xp = (nodes[n].x - Ax) * cos(anglePrincipalAxesRad) + (nodes[n].y - Ay) * sin(anglePrincipalAxesRad);
			nodes[n].yp = (nodes[n].y - Ay) * cos(anglePrincipalAxesRad) - (nodes[n].x - Ax) * sin(anglePrincipalAxesRad);

			nodes[n].x = nodes[n].xp;
			nodes[n].y = nodes[n].yp;
		}

	}
}


void Mesh::calculateSecondAreaMoments()
{
	Ix = 0;
	Iy = 0;
	Ixy = 0;

	for (int e = 0; e < nElem; e++)
	{
		Ix += elements[e].getSecondAreaMomentX(Ay);
		Iy += elements[e].getSecondAreaMomentY(Ax);
		Ixy += elements[e].getSecondAreaMomentProduct(Ax, Ay);
	}

	Imax = (Ix+Iy)/2 + sqrt( pow((Ix-Iy)/2, 2) + pow(Ixy, 2));

	calculateAnglePrincipalAxes();
}

// ********** Bending Moments **********

void Mesh::solveBendingMoments(){
	sigmaOfX.resize(nodes.size(),1);
	sigmaOfY.resize(nodes.size(),1);
	for(unsigned n=0; n<nodes.size(); n++){
		nodes[n].NofX = (nodes[n].y - Ay) / Ix;
		nodes[n].NofY = (nodes[n].x - Ax) / Iy;
		sigmaOfX(n,0) = (nodes[n].y - Ay) / Ix;
		sigmaOfY(n,0) = (nodes[n].x - Ax) / Iy;
	}

}

// ********** Torsional Moment **********

void Mesh::solveTorsion(){
	double cpu0 = clock();

	for(unsigned e=0; e<elements.size(); e++){
		elements[e].calculateElementLoadTorsion();
	}

	systemEquations(stiffness, load);
	constrainNode(stiffness, load, 0);

	double cpu1 = clock();

	std::cout << "built problem in " << cpu1 - cpu0 << std::endl;

	solveEquations(stiffness, displacement, load);

	double cpu2 = clock();

	printTemp2 = displacement;

	for(unsigned e=0; e<elements.size(); e++){
		elements[e].setElementValue(displacement, elements[e].elementDisplacement);
		elements[e].computeTau();
	}

	for(unsigned n=0; n<nodes.size(); n++){
		nodes[n].tau.resize(2, 1);
		nodes[n].tau << 0, 0;
		int num = nodes[n].depElem.size();
		for(int i=0; i<num; i++){

			int elem = nodes[n].depElem[i];
			int ieg = nodes[n].depIEG[i];

			nodes[n].tau += elements[elem].tau[ieg];
		}
		nodes[n].tau *= 1.0/num;
	}

	tauXZtorsion.resize(nodes.size(), 1);
	tauYZtorsion.resize(nodes.size(), 1);
	tauRtorsion.resize(nodes.size(), 1);
	for(unsigned n=0; n<nodes.size(); n++){
		tauXZtorsion(n, 0) = nodes[n].tau(0, 0);
		tauYZtorsion(n, 0) = nodes[n].tau(1, 0);
		tauRtorsion(n, 0) = sqrt(pow(tauXZtorsion(n, 0), 2) + pow(tauYZtorsion(n, 0), 2));
	}

	double cpu3 = clock();
	std::cout << "set values in " << cpu3 - cpu2 << std::endl;
}

// ********** Shear X **********

void Mesh::solveShearX(){
	for(unsigned e=0; e<elements.size(); e++){
		elements[e].calculateElementLoadShearX();
	}

	systemEquations(stiffness, load);
	constrainNode(stiffness, load, 0);
	solveEquations(stiffness, displacement, load);

	printTemp1 = displacement;

	for(unsigned e=0; e<elements.size(); e++){
		elements[e].setElementValue(displacement, elements[e].elementDisplacement);
		elements[e].computeTau();
		int a = elements[e].initialStrain.size();
	}

	for(unsigned n=0; n<nodes.size(); n++){
		nodes[n].tau.resize(2, 1);
		nodes[n].tau << 0, 0;
		int num = nodes[n].depElem.size();
		for(int i=0; i<num; i++){

			int elem = nodes[n].depElem[i];
			int ieg = nodes[n].depIEG[i];

			nodes[n].tau += elements[elem].tau[ieg];
		}
		nodes[n].tau *= 1.0/num;
	}

	tauXZshearX.resize(nodes.size(), 1);
	tauYZshearX.resize(nodes.size(), 1);
	tauRshearX.resize(nodes.size(), 1);
	for(unsigned n=0; n<nodes.size(); n++){
		tauXZshearX(n, 0) = nodes[n].tau(0, 0);
		tauYZshearX(n, 0) = nodes[n].tau(1, 0);
		tauRshearX(n, 0) = sqrt(pow(tauXZshearX(n, 0), 2) + pow(tauYZshearX(n, 0), 2));
	}
}

// ********** Shear X **********

void Mesh::solveShearY(){
	for(unsigned e=0; e<elements.size(); e++){
		elements[e].calculateElementLoadShearY();
	}

	systemEquations(stiffness, load);
	constrainNode(stiffness, load, 0);
	solveEquations(stiffness, displacement, load);

	for(unsigned e=0; e<elements.size(); e++){
		elements[e].setElementValue(displacement, elements[e].elementDisplacement);
		elements[e].computeTau();
		int a = elements[e].initialStrain.size();
	}

	for(unsigned n=0; n<nodes.size(); n++){
		nodes[n].tau.resize(2, 1);
		nodes[n].tau << 0, 0;
		int num = nodes[n].depElem.size();
		for(int i=0; i<num; i++){

			int elem = nodes[n].depElem[i];
			int ieg = nodes[n].depIEG[i];

			nodes[n].tau += elements[elem].tau[ieg];
		}
		nodes[n].tau *= 1.0/num;
	}

	tauXZshearY.resize(nodes.size(), 1);
	tauYZshearY.resize(nodes.size(), 1);
	tauRshearY.resize(nodes.size(), 1);
	for(unsigned n=0; n<nodes.size(); n++){
		tauXZshearY(n, 0) = nodes[n].tau(0, 0);
		tauYZshearY(n, 0) = nodes[n].tau(1, 0);
		tauRshearY(n, 0) = sqrt(pow(tauXZshearY(n, 0), 2) + pow(tauYZshearY(n, 0), 2));
	}
}

void Mesh::systemEquations(SparseMatrix<double> &K, MatrixXd &R){
	K.resize(nodes.size(), nodes.size());
	K.reserve((1/4) * nodes.size() * nodes.size());
	R = MatrixXd::Zero(nodes.size(), 1);

	for(unsigned e=0; e<elements.size(); e++){
		elements[e].calculateStiffnessMatrix();

		addLocalToGlobalSparse(K, e);
		addLocalToGlobalVector(R, e);
	}
	K.makeCompressed();
}

void Mesh::solveEquations(SparseMatrix<double> & A, MatrixXd &b, MatrixXd &C){
	std::cout << "Solving: LDLT" << std::endl;
	double time0 = clock();
	SimplicialLDLT<SparseMatrix<double>> LDLT(A);

	if(LDLT.info() != Success){
		std::cout << "\nFactorisation failed\n";
	}
	else{std::cout << "\nfactorized successfully\n";}

	b = LDLT.solve(C);

	if(LDLT.info() != Success){
		std::cout << "\nSolution failed\n";
	}
	else{std::cout << "\nSolved successfully\n";}

	double time1 = clock();

	std::cout << "solved in time = " << time1 - time0 << std::endl;

}

void Mesh::calculateAnglePrincipalAxes(){
	if(Ixy > -0.01 && Ixy < 0.01){
		anglePrincipalAxes = 0;
	}else{
	
		double t = (Ix-Imax)/Ixy;
		double angle = atan(t)*180/M_PI;

		if(angle <= 45 && angle >=-45){
			anglePrincipalAxes = angle;
		}

		if(angle > 45){
			anglePrincipalAxes = angle - 90;
		}

		if(angle < -45){
			anglePrincipalAxes = angle + 90;
		}
	}

	anglePrincipalAxesRad = anglePrincipalAxes * M_PI / 180;

}


double Mesh::getAnglePrincipalAxes()
{
	if(Ixy == 0){
		return 0;
	}
	
	double t = (Ix-Imax)/Ixy;
	double angle = atan(t)*180/M_PI;

	if(angle <= 45 && angle >=-45){
		return angle;
	}

	if(angle > 45){
		return angle - 90;
	}

	if(angle < -45){
		return angle + 90;
	}

	else{return angle;}


	//double t = -2*Ixy/(Ix-Iy);
	//return atan(t)*0.5*180/M_PI;
}

void Mesh::addtoGlobalRowMatrix(MatrixXd &global, MatrixXd local, std::vector<int> IEG)
{
	for (int i = 0; i < static_cast<int>(IEG.size()); i++)
	{
		global(IEG.at(i), 0) = global(IEG.at(i), 0) - local(i, 0);
	}
}

void Mesh::addToGlobal2DMatrix(SparseMatrix<double> &systemMatrix, MatrixXd local, std::vector<int> IEG)
{
	for (int x = 0; x  < local.cols(); x++)
	{
		int xn = IEG[x];
		for (int y = 0; y < local.rows(); y++)
		{
			int yn = IEG[y];
			addSubMatrix(systemMatrix, yn, xn, local(y,x));
		}
	}
}

void Mesh::addSubMatrix(SparseMatrix<double> &systemMatrix, int n1, int n2, double value) {
	systemMatrix.coeffRef(n1,n2) += value;
}

void Mesh::addLocalToGlobalSparse(SparseMatrix<double> &mat, int e){

	std::vector<int> IEG = elements[e].getIEG();

	for(int i=0; i<3; i++){
		for(int j=0; j<3; j++){
			mat.coeffRef(IEG[i], IEG[j]) += elements[e].elementStiffness(i, j);
		}
	}
}

void Mesh::addLocalToGlobalVector(MatrixXd &mat, int e){
	std::vector<int> IEG = elements[e].getIEG();

	if(e==0){
		mat.resize(nodes.size(), 1);
	}

	for(int i=0; i<3; i++){

		mat(IEG[i], 0) -= ( elements[e].elementLoad(i, 0));
	}
}

void Mesh::addLocalToGlobalVector(VectorXd &vec, int e){
	std::vector<int> IEG = elements[e].getIEG();

	VectorXd size = VectorXd::Zero(nodes.size());

	vec = size;

	for(int i=0; i<3; i++){
		vec(IEG[i]) = elements[e].elementLoad(i);
	}

}

void Mesh::addLocalToGlobalSparse(SparseMatrix<double> &mat){
	for(unsigned e=0; e<elements.size(); e++){
		
		std::vector<int> IEG = elements[e].getIEG();
		
		for(int i=0; i<3; i++){
			for(int j=0; j<3; j++){
				mat.coeffRef(IEG[i], IEG[j]) += elements[e].elementStiffness(i, j);
			}
		}
	}
}

void Mesh::clean(MatrixXd &matrix){
	for (int i = 0; i < matrix.cols(); i++)
	{
		for (int j = 0; j < matrix.rows(); j++)
		{
			if(abs(matrix(j,i)) < 0.001) {
				matrix(j,i) = 0;
			}
		}
	}
}

void Mesh::printr(int array[], int len){
	std::cout << std::endl << "Node:" << std::endl;
	for (int i = 0; i < len; i++)
	{
		std::cout << array[i];
	}
	std::cout << std::endl;
}

void Mesh::constrainNode(SparseMatrix<double> &K, MatrixXd &R, int nNumber)
{

	for(int i=0; i<K.cols(); i++){
		K.coeffRef(i, nNumber) = 0;
	}

	for(int i=0; i<K.rows(); i++){
		K.coeffRef(nNumber, i) = 0;
	}


	K.coeffRef(nNumber, nNumber) = 1;

	R(nNumber) = 0;
}

Node Mesh::getNodeAt(int n){
	return nodes[n];
}

Element Mesh::getElementAt(int e){
	return elements[e];
}


std::vector<int> Mesh::getNodesNearCentre(int nrNodesRequired)
{
	std::vector<int> constrainedNodes;
	std::vector<double> nodeDistances;
	
	double dx = 0;
	double dy = 0;
	for (int i = 0; i < nNode; i++)
	{
		dx = nodes[i].x - Ax;
		dy = nodes[i].y - Ay;
		nodeDistances.push_back(sqrt(dx*dx + dy*dy));
	}

	if(nNode*log10(nNode) + nrNodesRequired*(1 - nNode) >= 0)
	{
			selectConstrainedNodesByIteration(constrainedNodes, nodeDistances, nrNodesRequired, nNode);
	}
	else
	{
		//Should rather use quicksort, then select
		selectConstrainedNodesByIteration(constrainedNodes, nodeDistances, nrNodesRequired, nNode);
	}

	return constrainedNodes;
}

void Mesh::selectConstrainedNodesByIteration(std::vector<int> &constrainedNodes, std::vector<double> &nodeDistances, int nrNodesRequired, int nNode)
{	
	int nodesAssigned = 0;
	while(nrNodesRequired > nodesAssigned)
	{
		int selectedNode = 0;
		double lowestValue = nodeDistances[selectedNode];

		for (int i = 1; i < nNode; i++)
		{
			if(nodeDistances.at(i) < lowestValue)
			{
				lowestValue = lowestValue;
				selectedNode = i;
			}
		}
		constrainedNodes.push_back(selectedNode);
		nodeDistances.at(selectedNode) = UINT_MAX;
		nodesAssigned++;
	}
}

void Mesh::flush(){
	stiffness.setZero();
	stiffness.makeCompressed();

	load.setZero();
	displacement.setZero();

	for(unsigned e=0; e<elements.size(); e++){
		elements[e].elementStiffness.setZero();
		elements[e].elementLoad.setZero();
		elements[e].elementDisplacement.setZero();
		elements[e].tau.pop_back();
		elements[e].tau.pop_back();
		elements[e].tau.pop_back();
	}

	for(unsigned n=0; n<nodes.size(); n++){
		nodes[n].tau.setZero();
	}

	

}
