#include "Node.h"
#include "Element.h"
#include <vector>
#include <Eigen/Sparse>
#include <Eigen/SparseCholesky>
#include <Eigen/SparseLU>
#include <Windows.h>
#include <time.h>
#include <math.h>

#ifndef MESH
#define MESH

using namespace Eigen;

//! Mesh contains mesh data and has function to calculate and retrieve mesh properties.
class Mesh
{

public:
	EIGEN_MAKE_ALIGNED_OPERATOR_NEW		//!< Neccessary for std::vector of objects with Eigen members.

	int nElem;			//!< Number of elements
	int nNode;			//!< Number of nodes

	double Ax;		//!< Area centre for mesh in X-direction
	double Ay;		//!< Area centre for mesh in Y-direction
	double totalArea;		//!< Total area of the mesh
	double areaStiffness;	//!< Sum of all Ai * Ei

	double Ix; 		//!< Second area moment for mesh in X-direction
	double Iy; 		//!< Second area moment for mesh in Y-direction
	double Ixy;		//!< Second area moment for mesh in XY-direction
	double Imax;	//!< Maximum second area moment for mesh along principal axis
	double anglePrincipalAxes; //!< Stores the angle of the principal axes in degrees.
	double anglePrincipalAxesRad; //!< Stores the angle of the principal axes in radians.

	/*! Constructor computes the area, areacentre and the stiffness matrix
		\param elements List of all elements
		\param nodes List of all nodes
	*/
	Mesh(std::vector<Element>, int, std::vector<Node>, int, double);
	Mesh();

	/*! Calculates the resulting stresses from unit bending moments \n
		\f$ \sigma = dy / I_x \f$
		\f$ \sigma = dx / I_y \f$
	*/

	void solveBendingMoments();	

	MatrixXd sigmaAxial;	//!< Stores nodal stress from axial force.

	MatrixXd sigmaOfX;	//!< Stores all nodal stresses from moment about x-axis.
	MatrixXd sigmaOfY;	//!< Stores all nodal stresses from moment about y-axis.

	void solveTorsion();	//!< Calculates the shear stress distribution under torsional load.
	void solveShearX();		//!< Calculates the shear stress distribution under shear force along x-axis.
	void solveShearY();		//!< Calculates the shear stress distribution under shear force along y-axis.

	/*! 
		Sets an uniform E-module for the material over the whole mesh.
	*/
	void setEModule(double);

	/*!
		Function run by constructor. It runs through all elements calculating 
		element areas and element centres for each element.\n

		It also calculates total area of mesh, and area centre of the mesh using \n
		\f$ A_x = \sum{( \frac{dx_e*a_e}{A}) } \f$ and \f$ A_y = \sum{ (\frac{dy_e*a_e}{A}) } \f$
	*/


	void calculateMeshProperties();	

	/*!
		Transforms coordinates for computation purposes. \n
		\f$ x' = x-A_xcos(\alpha) + y-A_ysin(\alpha) \f$ \n
		\f$ x' = y-A_ycos(\alpha) - x-A_xsin(\alpha) \f$
	*/

	void transformCoordinates();

	/*!
		Loops through all elements calculating the distance between mesh- and element area centres, 
		and calculates the second moment of inertia for the mesh's x-, y- and xy-axis. \n
	*/
	void calculateSecondAreaMoments();

	/*! 
		Calculates the angle of one of the inertial axis. \n
		\f$ 	t = \frac{-2I_{xy}}{I_x - I_y} 	\f$ 
		\n
		\f$ 	\alpha = \frac{1}{2}\arctan(t)\frac{180}{\pi} \f$
		\return Angle of one inertial axis.
	*/
	void calculateAnglePrincipalAxes();
	double getAnglePrincipalAxes();

	void calculateLineCenterNodes();

//	double computeShearYCentre();

	/*! 
		Solves the eq \n
		\f$ Kr = R \f$, \n
		where K is the stiffness matrix, and r and R are vectors with displacements and loads in each node.
	*/
	void solveEquations(SparseMatrix<double> &A, MatrixXd &b, MatrixXd &C);
	void constrainNodesNearCentre(int);
	void constrainNode(SparseMatrix<double> &K, MatrixXd &R, int nNumber);

	Node getNodeAt(int n);
	Element getElementAt(int e);

	MatrixXd tauXZtorsion;	//!< Stores nodal \f$ \tau_{xz} \f$ from torsional load
	MatrixXd tauYZtorsion;	//!< Stores nodal \f$ \tau_{yz} \f$ from torsional load
	MatrixXd tauRtorsion;	//!< Stores nodal \f$ \tau_{R} \f$ from torsional load

	MatrixXd tauXZshearX;	//!< Stores nodal \f$ \tau_{xz} \f$ from shear load in x-direction
	MatrixXd tauYZshearX;	//!< Stores nodal \f$ \tau_{yz} \f$ from shear load in x-direction
	MatrixXd tauRshearX;	//!< Stores nodal \f$ \tau_{R} \f$ from shear load in x-direction

	MatrixXd tauXZshearY;	//!< Stores nodal \f$ \tau_{xz} \f$ from shear load in y-direction
	MatrixXd tauYZshearY;	//!< Stores nodal \f$ \tau_{yz} \f$ from shear load in y-direction
	MatrixXd tauRshearY;	//!< Stores nodal \f$ \tau_{R} \f$ from shear load in y-direction

	MatrixXd printTemp1;	//!< Used in testing to store desired value for printing
	MatrixXd printTemp2;	//!< Used in testing to store desired value for printing

private:
	std::vector<Element> elements;	//!< List of all elements in mesh
	std::vector<Node> nodes;		//!< List of all nodes in mesh

	double E;		//!< Youngs modulus for the material. Defined for every element
	double nu;		//!< Poisson's number
	double G;		//!< Shear modulus. Function of E and nu
	
	/*! 
		Only for use during visual inspection of a matrix
		\param matrix Matrix to be cleaned before printed to screen
	*/
	void clean(MatrixXd &);
	void printr(int[], int);
	
	// ********** Torsional Moment **********

	//SparseMatrix<double> torsionalStiffness;			//!< System stiffness matrix
	//MatrixXd torsionalLoad;			//!< System load matrix
	//MatrixXd torsionalDisplacement;			//!< System displacement matrix
	//
	//SparseMatrix<double> shearXStiffness;
	//MatrixXd shearXLoad;
	//MatrixXd shearXDisplacement;

	//SparseMatrix<double> shearYStiffness;
	//MatrixXd shearYLoad;
	//MatrixXd shearYDisplacement;

	SparseMatrix<double> stiffness;	//!< System stiffness matrix
	MatrixXd load;	//!< system load vector
	MatrixXd displacement;	//!< system displacement vector

	/*! 
		Computes the global stiffness matrix for the mesh \n
		Stiffness matrix for each element is found by \n
		\f$ k_e = t_e A_e B^T G B \f$,\n
		where \n

		\f$ G_v = \frac{0.5E}{1+v} \f$ \n
		\f$ G = \left| \begin{array}{ccc}
					G_v & 0 \\
					0 & G_v \end{array} \right| \f$ \n

		\f$ B = \left| \begin{array}{ccc}
				y_2-y_3 & y_3-y_1 & y_1-y_2 \\
				x_3-x_2 & x_1-x_3 & x_2-x_1 \end{array} \right| \f$ \n
		B matrix relates the strains to displacements for a triangle element.
	*/
	void systemEquations(SparseMatrix<double> &K, MatrixXd &R);

		/*!
		Adds element stiffness matrix to system matrix given an IEG matrix.
		\param MatrixXd Reference to local element stiffness matrix
		\param MatrixXd Reference to system matrix
		\param Integer	List of node connectivity for element
	*/
	void addToGlobal2DMatrix(SparseMatrix<double> &, MatrixXd, std::vector<int>);
	void addtoGlobalRowMatrix(MatrixXd &, MatrixXd, std::vector<int>);
	
	void addLocalToGlobalSparse(SparseMatrix<double> &mat, int e);	//!< Uses getIEG() to assign elemental values to corresponding global parameters
	void addLocalToGlobalSparse(SparseMatrix<double> &mat);	//!< Uses getIEG() to assign elemental values to corresponding global paramaters
	void addLocalToGlobalVector(MatrixXd &mat, int e);
	void addLocalToGlobalVector(VectorXd &vec, int e);

	/*!
		Increases value at a position in matrix by a given value.
		\param Matrix matrix to edit
		\param position i-position in matrix.
		\param position j-position in matrix.
		\param value Value added to position in matrix.
	*/
	void addSubMatrix(SparseMatrix<double> &, int, int, double);

	void selectConstrainedNodesByIteration(std::vector<int> &, std::vector<double> &, int, int);
	/*! 
		Finds a number of nodes in mesh near centre
		\param number Number of nodes to constrain
		\return A list of nodes near mesh centre
	*/
	std::vector<int> getNodesNearCentre(int);

	/*! 
		Sets all the system variables to zero. (Mesh::stiffness, Mesh::load, \n
		Element::elementStiffness, Element::elementLoad, Element::elementDisplacement, Element::tau, \n
		Node::tau)
	*/
	void flush(); 

};

#endif