#include "stdafx.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <vector>
#include <iostream>
#include <fstream>
#include "ResultWriter.h"

std::ofstream file;

ResultWriter::ResultWriter(std::string fileName, Mesh mesh)
{
	ResultWriter::fileName = fileName;
	ResultWriter::mesh = mesh;
	open();
}

void ResultWriter::writeMomentStress(std::vector<Node> nodes, int nLength)
{
	double Ix = mesh.Ix;
	double Iy = mesh.Iy;
	double Ixy = mesh.Ixy;

	double angle = mesh.getAnglePrincipalAxes();
	double tx = cos(angle);
	double ty = sin(angle);

	Vector2d principalAxisx(tx, ty);
	Vector2d principalAxisy(ty, -tx);

	Vector2d areaCentre(mesh.Ax, mesh.Ay);
	Vector2d nodePosition, distanceCentreToPoint;

	std::vector<double> datax;
	std::vector<double> datay;
	double dx, dy;
	for (int n = 0; n < nLength; ++n)
	{
		nodePosition << nodes[n].x , nodes[n].y;
		distanceCentreToPoint << areaCentre - nodePosition;

		dx = (distanceCentreToPoint - distanceCentreToPoint.dot(principalAxisx)*principalAxisx).norm();
		dy = (distanceCentreToPoint - distanceCentreToPoint.dot(principalAxisy)*principalAxisy).norm();

		datax.push_back(dx/Ixy);	//sigma = M/I * y;	(M = 1)
		datay.push_back(dy/Ixy);
	}

	writeScalarData("Sigma due to M_x", datax);
	writeScalarData("Sigma due to M_y", datay);
}

void ResultWriter::writeNormalStress()
{
	std::vector<double> data;
	for (int i = 0; i < mesh.nNode; i++)
	{
		data.push_back(1.0);
	}
	writeScalarData("Sigma due to N", 	data);
}

void ResultWriter::writeInertialAxes()
{
	writeSingleValue("InertialAxes", mesh.getAnglePrincipalAxes());
}

void ResultWriter::writeMeshAreaProperties()
{
	double areaX = mesh.Ax;
	double areaY = mesh.Ay;
	double totalArea = mesh.totalArea;
	
	double Ix = mesh.Ix;
	double Iy = mesh.Iy;
	double Ixy = mesh.Ixy;

//	double shearCentreX = mesh.computeShearXCentre();
//	double shearCentreY = mesh.computeShearYCentre();

	file << "#MeshProperties\n";
	file << "1\n";
	file << "8\n";

	file << areaX << std::endl;
	file << areaY << std::endl;
	file << totalArea << std::endl;

//	file << shearCentreX << std::endl;
//	file << shearCentreY << std::endl;
	
	file << Ix << std::endl;
	file << Iy << std::endl;
	file << Ixy << std::endl;

	file << "#EndMeshProperties\n";
}

void ResultWriter::writeSingleValue(std::string name, double value)
{
	file << "#" << name << std::endl;
	file << "1\n";
	file << "1\n";
	file << value << std::endl;
	file << "#End" << name << std::endl;

}

void ResultWriter::writeScalarData(std::string name, std::vector<double> data)
{
	file << "#" << name << std::endl;
	file << "1\n";
	file << data.size() << std::endl;
	int n = 0;
	for (double nodeData : data)
	{
		file << ++n << " " << nodeData << std::endl;
	}

	file << "#End" << name << std::endl;
}

void ResultWriter::writeBoundaryNodes(int nNodes, std::vector<int> data)
{
	file << "#BoundaryNodes\n";
	file << "2\n";
	file << nNodes << "\n";
	int n = 0, bn = 0;
	for (int nodeData : data)
	{
		while(n++ < nodeData)
		{
			file << n << ";" << 0 << ";" << 0 << std::endl;
		}
		file << n << ";" << 100 << ";" << 100 << std::endl;
	}
	file << "#EndBoundaryNodes\n";
}


bool ResultWriter::open()
{
	file.open(fileName);
	if (!file.is_open())
	{
		return false;
	}
	return true;
}

void ResultWriter::save()
{
	file.close();
}

void ResultWriter::writeVtkHeader(){
	file << "# vtk DataFile Version 4.0\n";
	file << fileName << "\n";
	file << "ASCII\n";
}

void ResultWriter::writeMeshTopology(Mesh &mesh){
	file << "DATASET POLYDATA\n";
	file << "POINTS " << mesh.nNode << " double\n";
	for(int n=0; n<mesh.nNode; n++){
		file << mesh.getNodeAt(n).x << ' ' << mesh.getNodeAt(n).y << ' ' << 0 << "\n";
	}
	file << "POLYGONS " << mesh.nElem << ' ' << mesh.nElem * 4 << "\n";
	for(int e=0; e<mesh.nElem; e++){
		file << "3 " << mesh.getElementAt(e).i << ' ' << mesh.getElementAt(e).j << ' ' << mesh.getElementAt(e).k << "\n";
	}
	file << "POINT_DATA " << mesh.nNode << "\n";
}

void ResultWriter::writeMeshTopology(int nNode, std::vector<Node> nodes, int nElem, std::vector<Element> elements){
	file << "DATASET POLYDATA\n";
	file << "POINTS " << nNode << " double\n";
	for(int n=0; n<nNode; n++){
		file << nodes[n].x << ' ' << nodes[n].y << ' ' << 0 << "\n";
	}
	file << "POLYGONS " << nElem << ' ' << nElem * 4 << "\n";
	for(int e=0; e<nElem; e++){
		file << "3 " << elements[e].i << ' ' << elements[e].j << ' ' << elements[e].k << "\n";
	}
	file << "POINT_DATA " << nNode << "\n";
}



void ResultWriter::writeMeshTopology(Mesh &mesh, MatrixXd data){
	file << "DATASET POLYDATA\n";
	file << "POINTS " << mesh.nNode << " double\n";
	for(int n=0; n<mesh.nNode; n++){
		file << mesh.getNodeAt(n).x << ' ' << mesh.getNodeAt(n).y << ' ' << data(n, 0) * 0.01 << "\n";
	}
	file << "POLYGONS " << mesh.nElem << ' ' << mesh.nElem * 4 << "\n";
	for(int e=0; e<mesh.nElem; e++){
		file << "3 " << mesh.getElementAt(e).i << ' ' << mesh.getElementAt(e).j << ' ' << mesh.getElementAt(e).k << "\n";
	}
	file << "POINT_DATA " << mesh.nNode << "\n";
}


void ResultWriter::writePointScalarData(std::string name, MatrixXd data){
	file << "SCALARS " << name << " double 1\n";
	file << "LOOKUP_TABLE default\n";
	for(int i=0; i<data.rows(); i++){
		for(int j=0; j<data.cols(); j++){
			file << data(i, j);
		}
		file << "\n";
	}

}

void ResultWriter::writePointVectorData(std::string name, MatrixXd data){
	file << "VECTORS " << name << " double\n";
	if(data.cols() == 1){
		for(int i=0; i<data.rows(); i++){
			file << 0 << " " << 0 << " " << data(i, 0) << "\n";
		}
	}
	else{
		for(int i=0; i<data.rows(); i++){
			for(int j=0; j<data.cols(); j++){
				file << data(i, j) << " ";
			}
			file << "\n";
		}
	}
}
