#include "stdafx.h"
#include "Node.h"
#include <vector>
#include "Mesh.h"
#include <Eigen/StdVector>

#ifndef RESULTWRITER
#define RESULTWRITER

//!Opens and writes results to file.

/*! 
	Resultwriter is responsible for opening and write data to file. It also contains definitions and the structure of results.
	Most of the calculations is done in Mesh and Solver, 
	and only calculations regarding mapping a force to a specific element or node performed in this class.
*/
class ResultWriter
{
public:
	ResultWriter(std::string, Mesh mesh);

	bool open();
	void save();
	
	/*! 
		Writes a result containing forces in each nodes given an uniform normal stress over the mesh. \n
		> Result tag "#Sigma due to N"
	*/
	void writeNormalStress();
	
	/*! 
		Calculate stresses caused by an unit moment about the inertal axis
		> Result tag "#Sigma due to M_x" \n
		> Result tag "#Sigma due to M_y"

		It first creates a unit vector along the inertial axis. 
		Using vector algebra to calculate the distance from the inertial axis to the each node. \n
		\f$ y = || (\vec{a - p}) - ((\vec{a - p}) \cdot \vec{n})\vec{n}|| \f$

		Force in a node is given by \n
		\f$ \sigma = \frac{M}{I}\cdot y \f$, \n 
		where y is the distance calculated above. \n
		
		The result written by this method is based on a unit moment(M = 1). \n

		\param Vector<Node> containing stresses in x,y direction for all nodes
		\param Integer Number of nodes in vector
		\param Mesh reference to mesh class containing the data
	*/
	void writeMomentStress(std::vector<Node>, int);
	
	/*!
		Writes the angle of one of the inertial axes to file. \n
		> Result tag "#InertialAxes"
	*/
	void writeInertialAxes();
	/*! 
		Writes the mesh properties to result file. \n
		> Result tag "#MeshProperties" \n

		_Properties written:_ \n
		+ Area centre,\n
		+ Total area,\n
		+ Shear centre,\n
		+ Moment of inertia-X,\n
		+ Moment of inertia-Y,\n
		+ Moment of inertia product
	*/
	void writeMeshAreaProperties();

	void writeScalarData(std::string, std::vector<double>);
	void writeSingleValue(std::string, double);
	void writeBoundaryNodes(int, std::vector<int>);
	//void writeSecondMomentArea(double, double, double);		//not in use

	// *************** VTK ***************
	
	void writeVtkHeader();
	void writeMeshTopology(Mesh &mesh, MatrixXd data);
	void writeMeshTopology(Mesh &mesh);
	void writeMeshTopology(int nNode, std::vector<Node>, int nElem, std::vector<Element>);
	void writePointScalarData(std::string name, MatrixXd data);
	void writePointVectorData(std::string name, MatrixXd data);
	

private:
	Mesh mesh;				//!< Reference to Mesh object containg mesh data
	std::string fileName;	//!< File name of file to write results 
};

#endif